/*!
 * \file MEMM.cpp
 * \brief MEMM abstract class implementation
 * \author Jean-Francois REY
 * \version 1.0
 * \date 23 Sept 2014
 */


#include "MEMM.hpp"

MEMM::MEMM() : dist_norm(0,1.0), unif(0,1.0), generator(static_cast<unsigned int>(std::time(0))), gauss(generator,dist_norm), accept(generator,unif)
{
  NPar = 0;
  Ns = 0;
  ParamFec = IndivFec = ParamDisp = NULL;
  pLoi = NULL;

}

MEMM::~MEMM()
{
  if(ParamFec)
  {
    ParamFec->close();
    delete ParamFec;
  }
  if(IndivFec)
  {
    IndivFec->close();
    delete IndivFec;
  }
  if(ParamDisp)
  {
    ParamDisp->close();
    delete ParamDisp;
  }
  if(pLoi) delete pLoi;

  fatherID.clear();

}

void MEMM::init(Parameters * p)
{

  if(p == NULL) return;

  // Load parents from file and fatherID
  loadParentFile(p);

  // Load seeds 
  loadSeeds(p);

  // Calcul distances
  calculDistances(p);

  // load allelic frequencies
  loadAllelicFreq(p);

  // calcul transition matrix
  calculTransitionMatrix(p);  

  calculWeightPollenDonor();

  loadPrior(p);
  loadParameters(p);

  for (int k=0; k<NPar; k++) {delete allParents[k];}
  for (int k=0; k<Ns; k++) {delete allSeeds[k];}
}

void MEMM::burnin()
{
    cout << "Burn in "<<Nburn<<" iterations ..." <<endl;
    if(Nburn) mcmc(Nburn,0);
    cout << "Burn in iterations : OK" <<endl<<endl;

}

void MEMM::run()
{
    cout << "Run " << Nstep <<" iterations" << endl;
    if(Nstep) mcmc(Nstep,Nthin);
    cout << "MCMC finished " << Nstep << " iterations"
    << endl << "You can visualize the results from R"<<endl;
}



/***************************** PROTECTED ***********************************/

void MEMM::loadParentFile(Parameters * p)
{
  //////////////////////////////////
  //    LOADING PARENTS           //
  //////////////////////////////////

  string name;
  char * temp;

  if(p == NULL) return;

  if(allParents.size() != 0) allParents.clear();

  cout<<endl<<"# Loading parents..."<<endl;
  temp = p->getValue(MEMM_PARENTS_FILE_NAME);
  ifstream parents(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_PARENTS_FILE_NUMBER_OF_LOCUS);
  Nl = atoi(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_PARENTS_FILE_CLASS_COV);
  Nql = atoi(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_PARENTS_FILE_QT_COV);
  Nqt = atoi(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_PARENTS_FILE_WEIGHT_VAR);
  Nw = atoi(temp);
  if(temp) delete []temp;

  NPar = 0;
  if(parents.good())
  {
    while(parents >> name)
    {
      cout<<name<<"...";
      allParents.push_back(new parent(name,parents,Nql,Nqt,Nl));
      NPar++;
    }
    for(int k=0; k<NPar; k++)
    {
      fatherID[(*allParents[k]).getName()]=k;
    }
    cout<<endl<< NPar << " parents loaded. " << std::endl;
  }
  else
  {
    cerr<<"#ERROR ! Can not read parents file !"<<endl<<"Continue..."<<endl;
  }
  parents.close();

}

void MEMM::loadSeeds(Parameters * p)
{

  int Nsl;
  string name;
  char * temp, *filename;

  if(p == NULL)return;

  cout<<endl<<"# Loading seeds..."<<endl;
  if(allSeeds.size() != 0) allSeeds.clear();

  filename = p->getValue(MEMM_OFFSPRING_FILE_NAME);
  ifstream seeds(filename);
  temp = p->getValue(MEMM_OFFSPRING_FILE_NUMBERS_OF_LOCUS);
  Nsl = atoi(temp);
  if(temp) delete []temp;

  Ns = 0;
  if(seeds.good())
  {
    while(seeds >> name)
    {
      cout<<name<<"...";
      allSeeds.push_back(new graine(name,seeds,Nsl));
      Ns++;
    }
    cout<<endl<<Ns<<" seeds loaded."<<endl;
    seeds.close();
  }
  else
  {
    cerr<<"#ERROR ! Can not read "<<filename<<" seeds file !"<<endl<<"Continue..."<<endl;
  }
  if(filename) delete []filename;

}

void MEMM::calculDistances(Parameters *p)
{
  //cerr<<"# Call of Default function MEMM::CalculDistances()\n  this fuction is not implemented"<<endl;
  if(allParents.size() != NPar || NPar == 0)
  {
    cerr<<"# ERROR : CalculDistance, number of parents incorrect"<<endl;
    return;
  }

  for(int i=DistMP.size()-1; i>=0 ; i--) DistMP[i].clear();
  DistMP.clear();
  DistMP.resize(NPar,vector<double>(NPar,0.));
  cout << endl << "# Computing distances from x-y coordinates of parents...";
  for (int m=0; m<NPar; m++)
    for (int p=0; p<NPar; p++)
      DistMP[m][p]=(*allParents[m]).dist(*allParents[p]);

    cout<<endl << "Computing distances : OK. " << endl;
}

void MEMM::loadAllelicFreq(Parameters * p)
{

  char * temp;
  int ntemp;
  double floattemp, stemp;//,a,ab;
  vector< std::vector<int> > AllSize (Nl);
  vector< std::map<int,double> > SizeFreq (Nl);
  Na.clear();
  Na.assign(Nl, 0);

  for(int i=AllFreq.size()-1; i>=0 ; i--) AllFreq[i].clear();
  AllFreq.clear();
  AllFreq.resize(Nl);

  for(int i=SizeAll.size()-1; i>=0 ; i--) SizeAll[i].clear();
  SizeAll.clear();
  SizeAll.resize(Nl);

  temp = p->getValue(MEMM_AF_FILE_MODE);
  if( strcmp("file",temp) == 0)
  {
    delete []temp;
    temp = p->getValue(MEMM_AF_FILE_NAME);
    ifstream fall(temp);

    cout << endl<<"# Loading allelic frequencies from file " <<temp<<" ..."<<endl;
    if(temp) delete []temp;
    for (int l=0; l<Nl; l++)
    {
      fall >> Na[l];
      stemp=0;
      for (int a=0; a<Na[l]; a++)
      {
        fall >> ntemp >> floattemp;
        stemp+=floattemp;
        AllSize[l].push_back(ntemp);
        AllFreq[l].push_back(floattemp);
        SizeAll[l][ntemp]=a;
      }
      cout << "Locus " << l << " : " << Na[l] << " alleles" <<endl;
      for (int a=0; a<Na[l]; a++)
      {
        AllFreq[l][a] /= stemp;
        SizeFreq[l][AllSize[l][a]]=AllFreq[l][a];
        std::cout << AllSize[l][a] << ":" << AllFreq[l][a] << " | ";
      }
      cout << endl;
    }
    cout << "Loading of allelic frequencies : OK. " <<endl;
    fall.close();
  }
  else
  {
    if(temp) delete []temp;
    int all;
    std::cout << endl << "# Computing allelic frequencies from adult trees..." << std::endl;
    for (int l=0; l<Nl; l++)
    {
      ntemp=0;
      for (int k=0; k<NPar; k++)
      {
        for (int m=0; m<2; m++)
        {
          all = (*allParents[k]).getGeno(l)._allele[m];
          if (all>-1)
          {
            if ( (SizeAll[l].find(all)) == SizeAll[l].end() )
            {
              Na[l]++;
              ntemp++;
              std::cout << Na[l] << "...";
              AllSize[l].push_back(all);
              AllFreq[l].push_back(1.);
              SizeAll[l][all]=Na[l]-1;
            }
            else {
              ntemp++;
              AllFreq[l][SizeAll[l][all]]++;
            }
          }
        }
      }
      std::cout << "Locus " << l << " : " << Na[l] << " alleles (" << ntemp << ")" <<std::endl;

      for (int a=0; a<Na[l]; a++)
      {
        AllFreq[l][a] /= ntemp;
        SizeFreq[l][AllSize[l][a]]=AllFreq[l][a];
        cout << AllSize[l][a] << ":" << AllFreq[l][a] << " | ";
      }
      cout << endl;
    }
    std::cout << "Estimation of allelic frequencies : OK. " <<std::endl << std::endl;
  }

}

void MEMM::calculTransitionMatrix(Parameters *p)
{

  cerr<<endl<<"# Call of Default function MEMM::CalculTransitionMatrix()\n  this fuction is not implemented"<<endl;

}

void MEMM::loadPrior(Parameters *p)
{
  char * temp, * temp2, *temp3;

  cout << endl << "Reading prior..."<<endl;
  temp = p->getValue(MEMM_GAMA_INIT);
  GamA = atof(temp);
  temp2 = p->getValue(MEMM_GAMA_MIN);
  mGamA = atof(temp2);
  temp3 = p->getValue(MEMM_GAMA_MAX); 
  MGamA = atof(temp3);
  cout << "GamA : "<< GamA << " [" <<mGamA <<"|"<< MGamA << "]"<<endl;
  if(temp) delete []temp;
  if(temp2) delete []temp2;
  if(temp3) delete []temp3;

  temp = p->getValue(MEMM_DELTA_INIT);
  Scale = atof(temp);
  temp2 = p->getValue(MEMM_DELTA_MIN);
  mScale = atof(temp2);
  temp3 = p->getValue(MEMM_DELTA_MAX);
  MScale = atof(temp3);
  cout << "Scale : "<< Scale << " [" <<mScale <<"|"<< MScale << "]"<<endl;
  if(temp) delete []temp;
  if(temp2) delete []temp2;
  if(temp3) delete []temp3;

  temp = p->getValue(MEMM_SHAPE_B_INIT);
  Shape = atof(temp);
  temp2 = p->getValue(MEMM_SHAPE_B_MIN);
  mShape = atof(temp2);
  temp3 = p->getValue(MEMM_SHAPE_B_MAX);
  MShape = atof(temp3);
  cout << "Shape : "<< Shape << " [" <<mShape <<"|"<< MShape << "]"<<endl;
  if(temp) delete []temp;
  if(temp2) delete []temp2;
  if(temp3) delete []temp3;
  
  temp = p->getValue(MEMM_MIG_RATE_M_INIT);
  Mig = atof(temp);
  temp2 = p->getValue(MEMM_MIG_RATE_M_MIN);
  mMig = atof(temp2);
  temp3 = p->getValue(MEMM_MIG_RATE_M_MAX);
  MMig = atof(temp3);
  cout << "Mig : "<< Mig << " [" <<mMig <<"|"<< MMig << "]"<<endl;
  if(temp) delete []temp;
  if(temp2) delete []temp2;
  if(temp3) delete []temp3;

  temp = p->getValue(MEMM_SELF_RATE_S_INIT);
  Self = atof(temp);
  temp2 = p->getValue(MEMM_SELF_RATE_S_MIN);
  mSelf = atof(temp2);
  temp3 = p->getValue(MEMM_SELF_RATE_S_MAX);
  MSelf = atof(temp3);
  cout << "Self : "<< Self << " [" <<mSelf <<"|"<< MSelf << "]"<<endl;
  if(temp) delete []temp;
  if(temp2) delete []temp2;
  if(temp3) delete []temp3;
  cout<<"...end reading prior"<<endl;
}


void MEMM::loadParameters(Parameters *p)
{
  char * temp;
  cout << endl << "Load parameters..."<<endl;

  temp = p->getValue(MEMM_SEED);
  if(atoi(temp) >= 0)
  {
    cout << "New seed : "<<temp<<endl;
    generator.seed(static_cast<long unsigned int>(atoi(temp)));
  }
  if(temp) delete []temp;

  temp = p->getValue(MEMM_IND_FEC_DIST);
  if( strcmp("LN",temp) == 0 ) pLoi = new MEMM_logNormal(gauss);
  else pLoi = new MEMM_gamma(GamA,accept);
  if(temp) delete []temp;

  temp = p->getValue(MEMM_BURNIN);
  Nburn = atoi(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_ITE);
  Nstep = atoi(temp);
  if(temp) delete []temp;
  temp = p->getValue(MEMM_THIN);
  Nthin = atoi(temp);
  if(temp) delete []temp;
  cout<<endl<< "Burnin "<< Nburn << " | Iteration "<<Nstep<<" | Thin "<<Nthin<<endl;
  

  cout<<"Ouput files :"<<endl;
  temp = p->getValue(MEMM_GAMA_FILE_NAME);
  ParamFec = new std::ofstream(temp, ofstream::out);
  if(ParamFec->good()) cout<<"\t-"<<temp<<endl;
  else cerr<<"#ERROR openning "<<temp<<" file"<<endl;
  if(temp) delete []temp;

  temp = p->getValue(MEMM_IND_FEC_FILE_NAME);
  IndivFec = new std::ofstream(temp, ofstream::out);
  if(IndivFec->good()) cout<<"\t-"<<temp<<endl;
  else cerr<<"#ERROR openning "<<temp<<" file"<<endl;
  if(temp) delete []temp;

  temp = p->getValue(MEMM_DISP_FILE_NAME);
  ParamDisp = new std::ofstream(temp, ofstream::out);
  if(ParamDisp->good()) cout<<"\t-"<<temp<<endl;
  else cerr<<"#ERROR openning "<<temp<<" file"<<endl;
  if(temp) delete []temp;

  cout << "... end reading default parameters"<<endl;
}

void MEMM::calculWeightPollenDonor()
{

  Poids.resize( NPar, 1.0 );

  cout <<endl << "Calculing pollen donor weight..."<<endl;
  for (int p=0; p<NPar; p++)
  {
    for (int k=0; k<Nw; k++)
      Poids[p]*=((*allParents[p])._covarquanti[k]);
    cout << Poids[p] << " | ";
  }
  cout<<endl;
}


/***************************** PRIVATE ***********************************/
